from itertools import cycle
import random

def delete_from_list(list_of_dicts, dict_item):
    for i in range(len(list_of_dicts)):
        if dict_item in list_of_dicts:
            list_of_dicts.remove(dict_item)
            break
    return list_of_dicts

class BaseData():
    def get_data(self):
        pass
    
    has_items = None
    
    def __init__(self):
        self.has_initialized  = True
        data = self.get_data()
        random.shuffle(data)
        self.set_data(data)

    def set_data(self, data):
        self.data = data
        self.cycled_data = cycle(self.data)
        self.has_items = len(data) > 0

    def get_one(self):
        if self.has_items:
            return next(self.cycled_data)


    def remove_data(self, item):
        print('Removing ', item)
        self.set_data(delete_from_list(self.data, item))

    def get_unique_random(self):
        if self.has_items:
            unique_items   = self.get_unique_items()
            return random.choice(unique_items)

    def get_unique_items(self):
        return [dict(s) for s in set(frozenset(d.items()) for d in self.data)]

    def get_unique_random_value(self):
        return next(iter(self.get_unique_random().values()))

    def get_hashed_value(self, value):
        unique_items   = self.get_unique_items()
        return unique_items[hash(value if value is not None else '_') % len(unique_items)]

    def get_hundred(self):
        return self.data


    def get_n(self, n):
        ls = []
        for i in range(n):
            ls.append(self.get_one())
        return ls


if __name__ == "__main__":
        
    class Test(BaseData):
        def get_data(self):
            if False:
            # if IS_PRODUCTION:
                N_US = 0
                N_GB = 0
                N_FR = 0
                N_NL = 100
                N_NO = 0
                N_CA = 0
                N_IN = 0
                # N_US = 50
                # N_GB = 10
                # N_FR = 10
                # N_NL = 10
                # N_NO = 10
                # N_CA = 10
                # N_IN = 0
            else:
                N_US = 1
                N_GB = 1
                N_FR = 1
                N_NL = 1
                N_NO = 1
                N_CA = 1
                N_IN = 1

            US = [{"country_code": "US"}] * N_US
            GB = [{"country_code": "GB"}] * N_GB
            FR = [{"country_code": "FR"}] * N_FR
            NL = [{"country_code": "NL"}] * N_NL
            NO = [{"country_code": "NO"}] * N_NO
            CA = [{"country_code": "CA"}] * N_CA
            IN = [{"country_code": "IN"}] * N_IN

            result = US+GB+FR+NL+NO+CA+IN
            return result


    TestInstance = Test() 
    print(TestInstance.data)
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    TestInstance.remove_data(TestInstance.get_one())
    print(TestInstance.data)



from urllib.error import ContentTooShortError
from http.client import RemoteDisconnected
from urllib.error import ContentTooShortError, URLError

from greatawesomeutils.lang import is_errors_instance

NETWORK_ERRORS = [RemoteDisconnected, URLError,
                  ConnectionAbortedError, ContentTooShortError,  BlockingIOError]


def is_error(errs):
    def fn(e):
        result, index = is_errors_instance(errs, e)
        return result
    return fn


is_network_error = is_error(NETWORK_ERRORS)



from .base_data import BaseData

class CountryCodes(BaseData):
    def get_data(self):
        if False:
        # if IS_PRODUCTION:
            N_US = 0
            N_GB = 0
            N_FR = 0
            N_NL = 100
            N_NO = 0
            N_CA = 0
            N_IN = 0
            # N_US = 50
            # N_GB = 10
            # N_FR = 10
            # N_NL = 10
            # N_NO = 10
            # N_CA = 10
            # N_IN = 0
        else:
            N_US = 0
            N_GB = 0
            N_FR = 0
            N_NL = 100
            N_NO = 0
            N_CA = 0
            N_IN = 0

        US = [{"country_code": "US"}] * N_US
        GB = [{"country_code": "GB"}] * N_GB
        FR = [{"country_code": "FR"}] * N_FR
        NL = [{"country_code": "NL"}] * N_NL
        NO = [{"country_code": "NO"}] * N_NO
        CA = [{"country_code": "CA"}] * N_CA
        IN = [{"country_code": "IN"}] * N_IN

        result = US+GB+FR+NL+NO+CA+IN
        return result


CountryCodesInstance = CountryCodes() 




from selenium.webdriver.common.keys import Keys
from time import sleep
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import  WebElement
from selenium.webdriver.common.keys import Keys
from datetime import datetime
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from greatawesomeutils.config import LOGS_PATH, SCREENSHOTS_PATH
from .relative_path import relative_path
from greatawesomeutils.config import WAIT_DURATION
from greatawesomeutils.lang import retry, write_file, write_html


def get_search_box(driver):
    m = driver.find_element(By.NAME, "q")
    return m


def search_for(driver, keyword):

    def doit():

        m = get_search_box(driver)
        # enter search text
        m.send_keys(keyword)
        sleep(1)
        # perform Google search with Keys.ENTER
        m.send_keys(Keys.ENTER)

    def run():
        doit()
        value = get_search_box(driver).get_attribute("value")
        if value != keyword:
            raise Exception(
                f"Expected keyword: {keyword} but was: {value}"
            )

    retry(run, 2, 2)


def get_direct(driver: WebDriver, link, src='Direct'):
    print(f"{src} Click: {link}")
    driver.get(link)


def get_twitter(driver: WebDriver, link):
    driver.get(link)

    # target="_blank"
    selector = "a.css-4rbku5.css-18t94o4.css-1dbjc4n.r-1loqt21.r-1ny4l3l.r-1udh08x.r-o7ynqc.r-6416eg.r-13qz1uu"

    el = get_element_or_none_by_selector(
        driver, selector, WAIT_DURATION)

    href = el.get_attribute("href")

    print(f"Twitter Click: {href}")

    # Prevent opening in new tab
    driver.execute_script(
        f'''document.querySelector("{selector}").removeAttribute('target')''')

    el.click()


def add_cookies(driver: WebDriver, cookies):
    for key in cookies:
        driver.add_cookie({"name": key, "value": cookies[key]})


def get_cookies(driver):
    all_cookies = driver.get_cookies()
    cookies_dict = {}
    for cookie in all_cookies:
        cookies_dict[cookie['name']] = cookie['value']
    return cookies_dict


def delete_all_site_data(driver: WebDriver):
    driver.delete_all_cookies()
    driver.execute_script("window.localStorage.clear();")
    driver.execute_script("window.sessionStorage.clear();")


SAVE_SCREENSHOT_ERROR_FOLDER = 'errors'
SAVE_SCREENSHOT_DEBUG_FOLDER = 'debug'
SAVE_SCREENSHOT_SUCCESS_FOLDER = 'success'


def get_html(driver):
    html = driver.execute_script("return document.body.innerHTML;")
    return html


def save_html_logs(driver, ID):
    html = get_html(driver)
    path = f'{LOGS_PATH}/{ID}.html'
    write_html(path, html)
    print('Written HTML ', path)


def save_screenshot(driver, ID, folder_type, tag=None):
    now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    try:
        # success
        tag = '-' if tag is None else f'-{tag}-'
        saving_screenshot_at = relative_path(
            f'{SCREENSHOTS_PATH}/{folder_type}/{ID}{tag}{now}.png', 2)
        driver.get_screenshot_as_file(
            saving_screenshot_at)
        print(f'Saved Screenshot at location: ..{saving_screenshot_at}')
    except:
        print('Failed to save screenshot')


def find_element_by_selector_with_wait(driver, xpath, wait):
    return WebDriverWait(driver, wait).until(
        EC.presence_of_element_located((By.CSS_SELECTOR, xpath)))


def get_element_or_none(driver, xpath, wait=None):
    try:
        if wait is None:
            return driver.find_element(By.XPATH, xpath)
        else:
            return WebDriverWait(driver, wait).until(
                EC.presence_of_element_located((By.XPATH, xpath)))
    except:
        return None


def get_element_or_none_by_text_contains(driver, text, wait=None):
    text = f'//*[contains(text(), "{text}")]'
    return get_element_or_none(driver, text, wait)


def get_element_or_none_by_text(driver, text, wait=None):
    text = f'//*[text()="{text}"]'
    
    return get_element_or_none(driver, text, wait)


def get_element_parent(element):
    return element.find_element(By.XPATH, "./..")


def get_element_or_none_by_selector(driver: WebDriver, selector, wait=None) -> WebElement:
    try:
        if wait is None:
            return driver.find_element(By.CSS_SELECTOR, selector)
        else:
            return WebDriverWait(driver, wait).until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, selector)))
    except:
        return None


def get_element_by_id(driver: WebDriver, id, wait=None):
    return get_element_or_none_by_selector(driver, f'[id="{id}"]', wait)



def get_elements_or_none_by_selector(driver: WebDriver, selector, wait=None):
    try:
        if wait is None:
            return driver.find_elements(By.CSS_SELECTOR, selector)
        else:
            WebDriverWait(driver, wait).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, selector)))

            return driver.find_elements(By.CSS_SELECTOR, selector)
    except:
        return None


def get_element_or_none_by_name(driver, selector, wait=None):
    try:
        if wait is None:
            return driver.find_element(By.NAME, selector)
        else:
            return WebDriverWait(driver, wait).until(
                EC.presence_of_element_located((By.NAME, selector)))
    except:
        return None


def print_body(driver):
    html = driver.find_element(
        By.XPATH, "/html/body").get_attribute('innerHTML')
    return html


def click(driver, xpath):
    driver.find_element(
        By.XPATH, xpath).click()


def assert_in_page(driver, page):
    assert (page in driver.current_url)


def js_click(driver, element):
    driver.execute_script("arguments[0].click();",  element)



from greatawesomeutils.config import *
from greatawesomeutils.lang import retry_if_is_error
import random
from bs4 import BeautifulSoup
import requests
import re
from greatawesomeutils.errors import NETWORK_ERRORS

def extract_links_from_html(html):
    soup = BeautifulSoup(html, features="html.parser")
    xs = []
    for a in soup.find_all('a', href=True):
        href = a['href']
        if 'http' in href:
            xs.append(href)

    return xs


def extract_links_from_text(text):
    result = re.search("(?P<url>https?://[^\s]+)", text).group("url")
    return [result]


API = 'https://www.1secmail.com/api/v1/'


def extractids(req):
    idList = []
    for i in req:
        for k, v in i.items():
            if k == 'id':
                mailId = v
                idList.append(mailId)
    return idList


# ["1secmail.com","1secmail.org","1secmail.net","bheps.com","dcctb.com","kzccv.com","qiott.com","wuuvo.com"]
domainList = None


def get_domains():
    global domainList
    if domainList is None:
        domainList = requests.get(
            "https://www.1secmail.com/api/v1/?action=getDomainList").json()
    return domainList


class TempMail():
    def get_domains():
        return get_domains()

    def generate_email(username):
        # ['1secmail.com', '1secmail.net', '1secmail.org']
        domain = random.choice(get_domains())

        email = f'{username}@{domain}'
        return email

    def extract(email):
        ls = email.split('@')
        login = ls[0]
        domain = ls[1]
        return login, domain

    def deleteMailbox(email):
        login, domain = TempMail.extract(email)
        url = 'https://www.1secmail.com/mailbox'
        data = {
            'action': 'deleteMailbox',
            'login': f'{login}',
            'domain': f'{domain}'
        }

        requests.post(url, data=data)

    def get_email_link(email):
        def run():
            login, domain = TempMail.extract(email)
            reqLink = f'{API}?action=getMessages&login={login}&domain={domain}'
            req = requests.get(reqLink).json()

            if len(req) == 0:
                print('No messages')
                assert False

            id = extractids(req)[-1]

            msgRead = f'{API}?action=readMessage&login={login}&domain={domain}&id={id}'
            req = requests.get(msgRead).json()

            html = req['htmlBody']
            
            if html == '':
                links = extract_links_from_text(req['textBody'])
            else:
                links = extract_links_from_html(html)
            has_no_links = len(links) == 0

            if has_no_links:
                print(req)

            return links[0]

        retry_if_is_error(
            run, NETWORK_ERRORS + [AssertionError], RETRY_ATTEMPTS_HIGH, WAIT_DURATION_VERY_SHORT)
        link = run()
        return link

    def get_body(email):
        def run():
            login, domain = TempMail.extract(email)
            reqLink = f'{API}?action=getMessages&login={login}&domain={domain}'
            req = requests.get(reqLink).json()

            if len(req) == 0:
                print('No messages')
                assert False

            id = extractids(req)[-1]

            msgRead = f'{API}?action=readMessage&login={login}&domain={domain}&id={id}'
            req = requests.get(msgRead).json()

            html = req['htmlBody']
            
            if html == '':
                return req['textBody']
            else:
                return  html

        retry_if_is_error(
            run, NETWORK_ERRORS + [AssertionError], 
            # 1
            RETRY_ATTEMPTS_HIGH
            , WAIT_DURATION_VERY_SHORT)
        data = run()
        return data

    def get_email_link_and_delete_mailbox(email):
        TempMail.get_email_link(email)
        TempMail.deleteMailbox(email)

if __name__ == '__main__':
    print('Running')
    email = TempMail.generate_email('temp')
    print(email)
    print(TempMail.get_email_link_and_delete_mailbox(email))



from greatawesomeutils.lang import get_valid_ip

def has_ip_changed(initial_ip):
    curr_ip = get_valid_ip()
    print("Current IP is: ", curr_ip)
    return initial_ip != curr_ip


def wait_till_ip_change():
    initial_ip = get_valid_ip()

    while True:
        if has_ip_changed(initial_ip):
            ip = get_valid_ip()
            print(f"IP changed from {initial_ip} to {ip}.")
            return ip


if __name__ == "__main__":
    print(has_ip_changed("192.0.0.1"))



import math
from greatawesomeutils.config import IS_DOCKER
from greatawesomeutils.config import RETRY_ATTEMPTS_HIGH
from greatawesomeutils.enums import *
from greatawesomeutils.errors import NETWORK_ERRORS
from greatawesomeutils.lang import country_code_to_time_zone, datetime_to_str, exit_with_failed_status, get_boolean_variable, get_time_in_utc, utc_to_timezoned_datetime, sleep_forever
from greatawesomeutils.lang import exit_with_failed_status, sleep_for_n_seconds
from greatawesomeutils.lang import retry_if_is_error, wait_for_enter
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver
from time import sleep
import json
import multiprocessing
import os
import os
import random
import requests
import signal
import time
import traceback
from .config import MASTER_API_URL
from .ip_utils import wait_till_ip_change
from .selenium import get_element_or_none, get_element_or_none_by_selector
from .selenium import get_element_or_none, get_element_or_none_by_selector
from .selenium import save_screenshot
from .selenium import save_screenshot
from .selenium import SAVE_SCREENSHOT_ERROR_FOLDER, SAVE_SCREENSHOT_SUCCESS_FOLDER
from .selenium import SAVE_SCREENSHOT_ERROR_FOLDER, SAVE_SCREENSHOT_SUCCESS_FOLDER, save_html_logs
from .selenium_driver import RANDOM, delete_cache
from .selenium_driver import delete_corrupted_files
from .selenium_driver import GOOGLE_DRIVER, UNDETECTED_DRIVER, create_driver_by_driver_type, delete_profile_path, MyChrome

exiting = False


def is_exiting():
    return exiting


def set_is_exiting():
    global exiting
    exiting = True


def ask_master_to_restart():
    data = [{
        "when": BaseTask.get_current_time_in_users_time_zone(),
        "data": {
            "task_id": BaseTask.task_id
        }
    }]

    r = requests.post(f"{BaseTask.MASTER_API}/tasks/restart/",
                      json=data)
    r.raise_for_status()
    # Wait For Process to be restarted by Master
    sleep_forever()


def execute_after_n_seconds(f, s):
    time.sleep(s)
    f()


class BotCaughtException(Exception):
    pass


PXM_ERROR_MSG = 'Bot has been caught by PerimeterX'


def is_bot_caught(driver):

    pmx = get_element_or_none(
        driver, "//*[text()='Please verify you are a human']")
    if pmx is not None:
        return True, PXM_ERROR_MSG

    clf = get_element_or_none_by_selector(driver, "#challenge-running")
    if clf is not None:
        return True, 'Bot has been caught by Cloudflare'

    return False, None


class EmailAlreadyExists(Exception):
    pass


class RetryException(Exception):
    pass


class RerunException(Exception):
    def __init__(self, after = None ):
      self.after = after


NULL = "NULL"


def has_user_id():
    return "user_id" in BaseTask.user


class BaseTask():
    data = json.loads(os.environ.get('data')) if os.environ.get('data') is not None else {} 
    user =  json.loads(os.environ.get('user')) if os.environ.get('user') is not None else {}
    task_type = os.environ.get('task_type')
    MODULE_NAME = os.environ.get('MODULE_NAME', '')

    task_id = os.environ.get('task_id')
    SKIP_IP_CHECK = get_boolean_variable('SKIP_IP_CHECK') if os.environ.get('SKIP_IP_CHECK') is not None else True
     
    is_debug = not IS_DOCKER

    MASTER_API = MASTER_API_URL

    def run(self, driver: MyChrome, data, user):
        pass

    
    def sign_up_success(self):
        r = requests.post(f"{BaseTask.MASTER_API}/tasks/sign-up-success/",
                          json={'task_id': BaseTask.task_id})
        r.raise_for_status()
        return True

    def save_failure_screenshot(student):
        save_screenshot(student.driver, BaseTask.user["user_id"],
                        SAVE_SCREENSHOT_SUCCESS_FOLDER, 'success')

    def assert_bot_has_not_been_caught(self):
        is_caught, message, = is_bot_caught(self.driver)

        if is_caught:
            raise BotCaughtException(message)

    def register_ip_for_task(ip):
        r = requests.post(f"{BaseTask.MASTER_API}/tasks/register-ip/",
                          json={'ip': ip, 'task_id': BaseTask.task_id})

        if r.status_code == 400:
            print('Ip already registered')
            return False

        r.raise_for_status()
        print('Registered Ip!')

        return True
    is_eager = False
    start_link = None
    
    register_ip = True

    def save_success_screenshot(self):
            save_screenshot(self.driver , BaseTask.user["user_id"],
                                SAVE_SCREENSHOT_SUCCESS_FOLDER, 'success')

    def begin_task(self):
            # Always use undetected in docker
        if IS_DOCKER:
            self.DRIVER_TYPE = UNDETECTED_DRIVER
        if BaseTask.SKIP_IP_CHECK:
            print(f'Skipping IP Check')
        else:
            ip = wait_till_ip_change()
            
            if self.register_ip:
                is_success = BaseTask.register_ip_for_task(ip)
                
                if not is_success:
                    ask_master_to_restart()
                # Skip IP check next time
            BaseTask.SKIP_IP_CHECK = True

        if not has_user_id():
            BaseTask.user["user_id"] = NULL

        is_sign_up_task = 'sign_up' in BaseTask.MODULE_NAME.lower()

        self.user = BaseTask.user
        self.data = BaseTask.data

        data = BaseTask.data
        user = BaseTask.user

        def do_run():
            driver = self.create_driver(self.is_eager)
            self.driver = driver
            try:
                if self.start_link is not None:
                    print('Organically getting ', self.start_link)
                    driver.organic_get(self.start_link)

                self.run(driver, data, user)
                self.save_success_screenshot()
                
                if is_sign_up_task:
                    self.sign_up_success()

                BaseTask.close(driver)
                print('Bot ran succesfully')

            except Exception as error:
                isReRun = isinstance(error, RerunException)
                
                def print_driver_url():
                    try:
                        driver_url = f'Driver Url: {driver.current_url}\n'
                        print(driver_url)
                    except:
                        pass

                if not isReRun:
                    traceback.print_exc()
                    print(error)
                
                if self.is_debug:
                    wait_for_enter('Waiting to Close')

                if not isReRun:
                    print_driver_url()
                    
                    save_screenshot(driver, user.get( "user_id", 'NULL'),
                                    SAVE_SCREENSHOT_ERROR_FOLDER, 'error')
                                    
                    save_html_logs(driver, user.get( "user_id", 'NULL'))

                BaseTask.close(driver)

                if isinstance(error, BotCaughtException):
                    print('Deletig Profile')
                    if user.get( "user_id") is not None:
                        delete_profile_path(BaseTask.user["user_id"])
                    ask_master_to_restart()
                elif isinstance(error, EmailAlreadyExists):
                    print('Deletig Profile')
                    if user.get( "user_id") is not None:
                        delete_profile_path(BaseTask.user["user_id"])
                    BaseTask.change_email()
                    ask_master_to_restart()

                if isinstance(error, RetryException):
                    ask_master_to_restart()
                if isReRun:
                    if self.retries < self.max_retries:
                        self.retries+=1
                        if error.after is not None:
                            sleep_for_n_seconds(error.after)

                        print('Bot is ReRunning')
                        do_run()
                    
                else:
                    exit_with_failed_status()
        do_run()

    retries = 0
    max_retries = math.inf  
    
    def change_email():
        r = requests.patch(f"{BaseTask.MASTER_API}/tasks/change-email/",
                           json={'task_id': BaseTask.task_id})
        r.raise_for_status()
        return True

    DRIVER_TYPE = GOOGLE_DRIVER


    def has_user_id(self):
        return  BaseTask.user.get("user_id") != NULL and BaseTask.user.get("user_id") is not None
    FORCE_VERSION = None

    def create_driver(self, is_eager):

        def do_create_driver():
            def run():

                driver_type = self.DRIVER_TYPE
                return create_driver_by_driver_type(
                    driver_type,
                    BaseTask.user.get("user_id") if self.has_user_id() else None,
                    BaseTask.user.get("window_size", RANDOM),
                    BaseTask.user.get("user_agent", RANDOM),
                    is_eager,
                    self.FORCE_VERSION)

            driver = retry_if_is_error(
                run, NETWORK_ERRORS + [(WebDriverException, lambda: delete_corrupted_files(BaseTask.user["user_id"]) if self.has_user_id() else None )], RETRY_ATTEMPTS_HIGH)

            def call_quit(signalNumber, frame):
                if is_exiting():
                    print('Already Exiting')
                    return

                set_is_exiting()

                print('Force Exit Detected. Exiting Gracefully')
                try:
                    BaseTask.close(driver)
                except Exception:
                    print('Someting went wrong closing Driver')
                finally:
                    sleep_for_n_seconds(2)
                    exit_with_failed_status()

            signal.signal(signal.SIGINT, call_quit)
            signal.signal(signal.SIGTERM, call_quit)
            return driver

        # Basically ask master to restart if failed to start after 10 minutes
        thread = multiprocessing.Process(
            target=execute_after_n_seconds, args=(ask_master_to_restart, 600))
        thread.start()
        driver = do_create_driver()
        thread.terminate()
        return driver

    def close(driver):
        delete_cache(driver)

        print("Closing Driver")
        driver.close()
        print("Closed Driver")
        # Sleep Time Needed for Chrome User Profiles To Be Saved Properly
        sleep(2)

    def get_current_time_in_users_time_zone():
        result = datetime_to_str(utc_to_timezoned_datetime(
            get_time_in_utc(), country_code_to_time_zone(BaseTask.user["country_code"])))

        return result



def sleep_for_word_length_type(s):

    def count_words(s):
        return len(s.split(' '))

    word_len = count_words(s)

    if word_len == 1:
        sleep_for_n_seconds(random.randint(4, 7))
    else:
        start = 5 + word_len
        end = 5 + word_len * 2
        print(start, end)
        sleep_for_n_seconds(random.randint(int(start), int(end)))



from urllib.request import urlopen
import json
from .lang import get_valid_ip


def get_country():
    IP = get_valid_ip()
    url = 'http://ipinfo.io/' + IP + '/json'
    response = urlopen(url)
    data = json.load(response)
    # city = data['city']
    # region = data['region']
    country = data['country']
    # location = data['loc']
    # org = data['org']
    return country



import os
IS_PRODUCTION = os.environ.get("ENV") == "production"
IS_DEVELOPMENT = not IS_PRODUCTION
IS_DIGITAL_OCEAN = os.environ.get("IS_DIGITAL_OCEAN") == 'True'


def is_docker():
    path = '/proc/self/cgroup'

    return (
        os.path.exists('/.dockerenv') or
        os.path.isfile(path) and any('docker' in line for line in open(path))
        or os.environ.get('KUBERNETES_SERVICE_HOST') is not None
    )


IS_DOCKER = is_docker()

TARGET_DIR = "production" if IS_PRODUCTION else "test"

TARGET_DIR_PARENT = 'digital-ocean-app-data' if IS_DIGITAL_OCEAN else 'app-data'
BASE_PATH = f'{TARGET_DIR_PARENT}/{TARGET_DIR}'
PROFILES_PATH = f'{BASE_PATH}/profiles'
SCREENSHOTS_PATH = f'{BASE_PATH}'

LOGS_PATH = f'../{BASE_PATH}/debug'

DB_PATH = f'{BASE_PATH}/app-data.db'

DOCKER_COMPOSE_YAML = f'docker-compose.yaml'
DOCKER_COMPOSE_YAML_FOR_VIDEOS = f'composes'

WAIT_DURATION_LONG = 180
WAIT_DURATION = 20
WAIT_DURATION_SHORT = 10
WAIT_DURATION_VERY_SHORT = 5
WAIT_DURATION_VERY_VERY_SHORT = 2


RETRY_ATTEMPTS_LOW = 3
RETRY_ATTEMPTS_HIGH = 5

SECONDS_IN_MINUTES = 60

MASTER_API_URL = "http://master-srv:4000" if IS_DOCKER else "http://localhost:4000"



from .base_data import BaseData
from greatawesomeutils.lang import *

class UserAgents(BaseData):
    def get_data(self):
        
        if is_windows():
            _106_AGENT = {"user_agent":        "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.37"}
            _105_AGENT = {"user_agent": "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36"}
            _104_2_AGENT = {"user_agent":  "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.102 Safari/537.36"}
            _104_1_AGENT = {"user_agent": "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.101 Safari/537.36"}
            _103_AGENT = {"user_agent":  "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36"}
            _101_AGENT = {"user_agent":  "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951 Safari/537.36"}
            _100_AGENT = {"user_agent":    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896 Safari/537.36"}
            _99_AGENT = {"user_agent":   "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844 Safari/537.36"}
            _98_AGENT = {"user_agent":    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758 Safari/537.36"}
            _97_AGENT = {"user_agent":    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692 Safari/537.36"}
            _96_AGENT = {"user_agent":    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664 Safari/537.36"}
            _95_AGENT = {"user_agent":    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638 Safari/537.36"}
            _94_AGENT = {"user_agent":   "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606 Safari/537.36"}
        else:
            _106_AGENT = {"user_agent":        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.37"}
            _105_AGENT = {"user_agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36"}
            _104_2_AGENT = {"user_agent":  "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.102 Safari/537.36"}
            _104_1_AGENT = {"user_agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.101 Safari/537.36"}
            _103_AGENT = {"user_agent":  "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36"}
            _101_AGENT = {"user_agent":  "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951 Safari/537.36"}
            _100_AGENT = {"user_agent":    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896 Safari/537.36"}
            _99_AGENT = {"user_agent":   "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844 Safari/537.36"}
            _98_AGENT = {"user_agent":    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758 Safari/537.36"}
            _97_AGENT = {"user_agent":    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692 Safari/537.36"}
            _96_AGENT = {"user_agent":    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664 Safari/537.36"}
            _95_AGENT = {"user_agent":    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638 Safari/537.36"}
            _94_AGENT = {"user_agent":   "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606 Safari/537.36"}


        # Versions
        V_106 = 37
        V_105 = 42
        V_104_1 = 1
        V_104_2 = 1
        V_103 = 2
        V_101 = 1
        V_99 = 10
        V_100 = 1
        V_98 = 1
        V_97 = 1
        V_96 = 1
        V_95 = 1
        V_94 = 1

        _106 = [_106_AGENT] * V_106
        _105 = [
            _105_AGENT] * V_105
        _104_1 = [
            _104_1_AGENT] * V_104_1
        _104_2 = [
            _104_2_AGENT] * V_104_2
        _103 = [
            _103_AGENT] * V_103
        _101 = [
            _101_AGENT] * V_101
        _99 = [
            _99_AGENT] * V_99
        _100 = [
            _100_AGENT] * V_100
        _98 = [
            _98_AGENT] * V_98
        _97 = [
            _97_AGENT] * V_97
        _96 = [
            _96_AGENT] * V_96
        _95 = [
            _95_AGENT] * V_95
        _94 = [_94_AGENT] * V_94

        result = _106 + _105 + _104_1 + _104_2 + _103 + \
            _101 + _99 + _100 + _98 + _97 + _96 + _95 + _94

        return result

UserAgentsInstance = UserAgents() 
if __name__ == '__main__':
    print(UserAgentsInstance.get_hashed_value('a'))
    print(UserAgentsInstance.get_hashed_value('a'))
    print(UserAgentsInstance.get_hashed_value('b'))



from .selenium import *
from greatawesomeutils.config import IS_DOCKER
from greatawesomeutils.config import PROFILES_PATH
from greatawesomeutils.lang import silentremove
from greatawesomeutils.user_agents import *
from greatawesomeutils.window_sizes import *
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as GoogleChromeOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from time import sleep
from undetected_chromedriver.v2 import Chrome, ChromeOptions
from .local_storage import LocalStorage
from .relative_path import relative_path
import shutil
from selenium.webdriver.common.action_chains import ActionChains
import pydash


class DriverUtils():
    def get_with_wait(self, link):
        self.get(link)
        # Wait for dom to be interactable
        sleep(2)

    def get_by_current_page_referrer(self, link, wait=2):

# selenium.common.exceptions.WebDriverException
        self.execute_script(f"""
                window.location.href = "{link}";
            """)
        if wait is not None and wait != 0:
            sleep(wait)

    def js_click(self, element):
        return js_click(self, element)


    def get_element_or_none_by_selector(self, selector, wait=None):
        return get_element_or_none_by_selector(self, selector, wait)

#     await new Promise((resolve) => {
#       var totalHeight = 0;
#       var distance = 100;
#       let count = 0;

#       var timer = setInterval(() => {
#         count++  
#         var scrollHeight = document.body.scrollHeight;
#           window.scrollBy(0, distance);
#           totalHeight += distance;
#           if (count >=5 ){
#             clearInterval(timer);
#             return resolve();

#           }else{

#           if(totalHeight >= scrollHeight - window.innerHeight){
#               clearInterval(timer);
#               resolve();
#           }}
#       }, 100);
#   });
# 
    def scroll_some(self):
        self.execute_script(""" 
window.scrollBy(0, 10000);
""")

    def scroll_element(self, element):

        is_till_end = self.execute_script("return arguments[0].scrollTop === (arguments[0].scrollHeight - arguments[0].offsetHeight)", element)

        if is_till_end:
            return False
        else:
            
            # self.execute_script("console.log(arguments[0])", element)
            # act = ActionChains(self)
            # act.scroll_from_origin(element, 0, -h).perform()

            # wait_for_enter()

            self.execute_script("arguments[0].scrollBy(0, 10000)", element)
            return True

# self.execute_script("""
# const el = arguments[0]
# const isTillEnd = 
# if (isTillEnd) {
#     return false;    
# } else {
#     el.scrollBy(0, 10000); 
#     return true;
# }
# """, element)

    def get_element_by_id(self, selector, wait=None):
        return get_element_by_id(self, selector, wait)

    def is_in_page(self, target, wait=None, raiseException=False):
        
        def check_page(driver, target):
            if isinstance(target, str):
                return target in driver.current_url
            else: 
                return pydash.some(target, lambda x: x in driver.current_url)
        
        if wait is None:
            return check_page(self, target)
        else:
            time = 0
            while time < wait:
                if check_page(self, target):
                    return True

                sleep_time = 0.2
                time += sleep_time
                sleep(sleep_time)
                
        if raiseException:
            raise Exception(f"Page {target} not found")
        return False

    def get_element_text(self, element):
        return element.get_attribute('innerHTML')

    def get_element_or_none_by_text_contains(self, text, wait=None):
        return get_element_or_none_by_text_contains(self, text, wait)

    def get_element_or_none_by_text(self, text, wait=None):
        return get_element_or_none_by_text(self, text, wait)

    def get_element_or_none_by_name(self, selector, wait=None):
        return get_element_or_none_by_name(self, selector, wait)
   
    def get_element_or_none(self, path, wait=None):
        return get_element_or_none(self, path, wait)

    def get_html(self):
        return get_html(self)

    def wait_till_element_to_disappear_by_selector(self,  selector,  wait):
        try:
            WebDriverWait(driver, wait).until(
                EC.invisibility_of_element_located((By.CSS_SELECTOR, selector)))
            return True
        except:
            return False

    def get_elements_or_none_by_selector(self, selector, wait=None):
        return get_elements_or_none_by_selector(self, selector, wait)

    def extract_links(self, elements):
        def extract_link(el):
            return el.get_attribute("href")

        return list(map(extract_link, elements))

    def extract_links_starting_with(self, start, wait=None):
        els = self.get_elements_or_none_by_selector("a", wait)

        links = self.extract_links(els)

        def is_not_none(link):
            return link is not None

        def is_starts_with(link):
            return link.startswith(start)

        return list(filter(is_starts_with, filter(is_not_none, links)))

    def print_body(self):
        return print_body(self)

    def organic_get_via(self, link, via, wait=2):
        def visit(driver):
            driver.get(via)

        def run(driver, link):
            visit(driver)
            driver.get_by_current_page_referrer(link, wait)

        run(self, link)

    def organic_get(self, link, wait=2):
        self.get_google()
        self.get_by_current_page_referrer(link, wait)

    def get_google(self):
        self.get("https://www.google.com/")
        self.get_element_or_none_by_selector('input[role="combobox"]', WAIT_DURATION)


    def get_innerhtml(self, el):
        return el.get_attribute("innerHTML")

    def get_cookies_as_dict(self):
        all_cookies = self.get_cookies()
        cookies_dict = {}
        for cookie in all_cookies:
            cookies_dict[cookie['name']] = cookie['value']
        return cookies_dict

    def delete_all_site_data(self):
        self.delete_all_cookies()
        self.execute_script("window.localStorage.clear();")
        self.execute_script("window.sessionStorage.clear();")

    def add_cookies(self, cookies):
        for key in cookies:
            self.add_cookie({"name": key, "value": cookies[key]})

    def get_local_storage(self):
        storage = LocalStorage(self)
        return storage.items()

    def add_local_storage(self, local_storage):
        storage = LocalStorage(self)
        for key in local_storage:
            storage.set(key, local_storage[key])

    def get_all_site_data(self):
        cookies = self.get_cookies_as_dict()
        local_storage = self.get_local_storage()

        return {"cookies": cookies, "local_storage": local_storage}

    def add_all_site_data(self, site_data):
        cookies = site_data["cookies"]
        local_storage = site_data["local_storage"]
        self.add_cookies(cookies)
        self.add_local_storage(local_storage)


class MyChrome(Chrome, DriverUtils):
    pass

class MyGoogleChrome(webdriver.Chrome, DriverUtils):
    # def aa(self):
    #     self.switch_to.default_content
    #     pass

    pass


class MyFirefox(webdriver.Firefox, DriverUtils):
    pass


def delete_cache(driver):
    print('Deleting Cache')
    driver.command_executor._commands['SEND_COMMAND'] = (
        'POST', '/session/$sessionId/chromium/send_command'
    )
    driver.execute('SEND_COMMAND', dict(
        cmd='Network.clearBrowserCache', params={}))


def add_useragent(options, user_agent):
    options.add_argument(f'--user-agent={user_agent}')


def create_profile_path(user_id):
    PATH = f'{PROFILES_PATH}/{user_id}'
    path = relative_path(PATH, 2)
    return path


def delete_corrupted_files(user_id):
    is_success = silentremove(
        f'{create_profile_path(user_id)}/SingletonCookie')
    silentremove(f'{create_profile_path(user_id)}/SingletonSocket')
    silentremove(f'{create_profile_path(user_id)}/SingletonLock')

    if is_success:
        print('Fixed Profile by deleting Corrupted Files')
    else:
        print('No Corrupted Profiles Found')


def delete_profile_path(user_id):
    path = create_profile_path(user_id)
    shutil.rmtree(path, ignore_errors=True)


RANDOM = "RANDOM"
HASHED = "HASHED"


def add_essential_options(options, user_id, window_size, user_agent):
    options.add_argument("--start-maximized")
    
    if window_size is not None:
        if window_size == RANDOM:
            window_size = WindowSizesInstance.get_unique_random_value()
        elif window_size == HASHED:
            window_size = WindowSizesInstance.get_hashed_value(user_agent)

        print("window-size: ", window_size)
        options.add_argument(f"--window-size={window_size}")
#
    if user_agent is not None:
        if user_agent == RANDOM:
            user_agent = UserAgentsInstance.get_unique_random_value()
        elif user_agent == HASHED:
            user_agent = UserAgentsInstance.get_hashed_value(user_agent)

        print("user_agent: ", user_agent)
        add_useragent(options, user_agent)

    has_user = user_id is not None
    if has_user:
        path = create_profile_path(user_id)
        print("user-data-dir: ", path)
        user_data_path = f"--user-data-dir={path}"
        options.add_argument(user_data_path)
    else:
        print("No User was Given")

    return options


def get_eager_startegy():

    caps = DesiredCapabilities().CHROME
    # caps["pageLoadStrategy"] = "normal"  #  Waits for full page load
    caps["pageLoadStrategy"] = "none"   # Do not wait for full page load
    return caps


UNDETECTED_DRIVER = 'undetected_chromedriver'
FIREFOX_DRIVER = 'firefox_driver'
GOOGLE_DRIVER = 'google_driver'


def hide_automation_flags(options):
    options.add_argument('--disable-blink-features=AutomationControlled')
    options.add_argument("--disable-blink-features")

    options.add_experimental_option(
        "excludeSwitches", ["enable-automation"])
    options.add_experimental_option('useAutomationExtension', False)

    # New Options
    options.add_argument("--ignore-certificate-errors")
    options.add_argument('--no-sandbox')
    options.add_argument("--disable-extensions")



def create_driver_by_driver_type(driver_type, user_id, window_size, user_agent, is_eager, FORCE_VERSION):

    if FIREFOX_DRIVER == driver_type:
        # TODO: ADD SUPPORT FOR PARAMS
        options = FirefoxOptions()
        options.headless = False

        options.binary_location = r'C:\Program Files\Mozilla Firefox\firefox.exe'

        has_user = user_id is not None
        if has_user:
            path = create_profile_path(user_id)
            options.set_preference('profile', path)
            print('SET PROFILES_PATH: ', path)

        driver = MyFirefox(
            options=options,
            # firefox_profile=add_tor(),
            # capabilities=create_proxy_capability('130.41.55.190:8080'),
            executable_path=relative_path("drivers/geckodriver.exe", 1),
            #  proxy=create_proxy('130.41.55.190:8080')
        )

    else:
        is_undetected = driver_type == UNDETECTED_DRIVER
        options = ChromeOptions() if is_undetected else GoogleChromeOptions()

        if IS_DOCKER:
            print("Running in Docker, So adding sandbox arguments")
            options.arguments.extend(
                ["--no-sandbox", "--disable-setuid-sandbox"])
        else:
            print("Not Running in Docker, So not adding sandbox arguments")

        add_essential_options(options, user_id, window_size, user_agent)

        if is_eager:
            desired_capabilities = get_eager_startegy()
            print('Driver will not wait for page to load fully')
        else:
            desired_capabilities = None
        print("Creating Driver")
        
        version = None
        
        if IS_DOCKER:
            version = 106
        elif FORCE_VERSION is not None: 
            version = FORCE_VERSION
        else: 
            version = 108

        if is_undetected:

            driver = MyChrome(desired_capabilities=desired_capabilities,
                              options=options, 
                              version_main=version
                            )
        else:
            # options.add_experimental_option("prefs",  {"profile.managed_default_content_settings.images": 2})
            hide_automation_flags(options)
            
            # CAPTCHA
            options.arguments.extend(
                ["--disable-web-security", "--disable-site-isolation-trials", "--disable-application-cache"])

            path = relative_path("drivers/chromedriver.exe", 2)
            print(f'driver path: {path}' )
            driver = MyGoogleChrome(
                desired_capabilities=desired_capabilities,
                chrome_options=options,
                executable_path=path,
            )
        print("Created Driver")

    return driver


def create_undetected_driver(user_id, window_size, user_agent, is_eager=False):
    return create_driver_by_driver_type(UNDETECTED_DRIVER, user_id, window_size, user_agent, is_eager)


def create_chrome_driver(user_id, window_size, user_agent, is_eager=False):
    return create_driver_by_driver_type(GOOGLE_DRIVER, user_id, window_size, user_agent, is_eager)


if __name__ == "__main__":
    # run using python -m .selenium_driver
    driver = create_chrome_driver()
    # visit_google(driver)
    # el = find_element_by_selector_with_wait(driver, "div.denfrfrgkbrg", 10)
    get_element_or_none_by_selector(driver, 'a.pHiOh').click()

    print(driver.current_url)
    # driver.find_element(By.NAME, "q")
    driver.close()
    driver.quit()
